#  tests for py-mxnet-1.5.0-py36h5ed9918_0 (this is a generated file);
print('===== testing package: py-mxnet-1.5.0-py36h5ed9918_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import mxnet as mx
a = mx.nd.ones((2, 3))
b = a * 2 + 1
b.asnumpy()
#  --- run_test.py (end) ---

print('===== py-mxnet-1.5.0-py36h5ed9918_0 OK =====');
print("import: 'mxnet'")
import mxnet

